import logging
import typing

from collections import OrderedDict
from typing import Any, Dict, List, Optional
from uuid import UUID

import six

from six import iteritems

from dbnd._core.configuration.config_readers import parse_and_build_config_store
from dbnd._core.configuration.config_value import ConfigValuePriority
from dbnd._core.configuration.dbnd_config import config
from dbnd._core.constants import RESULT_PARAM
from dbnd._core.parameter.parameter_builder import ParameterFactory
from dbnd._core.parameter.parameter_definition import (
    ParameterDefinition,
    _ParameterKind,
)
from dbnd._core.parameter.parameter_value import Parameters
from dbnd._core.task_build.decorated_callable_param_builder import (
    DecoratedCallableParamBuilder,
)
from dbnd._core.task_build.task_passport import TaskPassport
from dbnd._core.task_build.task_results import FuncResultParameter
from dbnd._core.task_build.task_signature import user_friendly_signature
from dbnd._core.task_build.task_source_code import TaskSourceCode
from dbnd._core.utils.basics.nothing import is_defined
from dbnd._core.utils.structures import combine_mappings
from dbnd._core.utils.uid_utils import get_uuid


if typing.TYPE_CHECKING:
    from dbnd._core.task_build.task_decorator import TaskDecorator

logger = logging.getLogger(__name__)


def _ordered_params(x):
    return OrderedDict(sorted(x.items(), key=lambda y: y[1].parameter_id))


class TaskDefinition(object):
    """
    TaskDefinition contains all the information gathered and calculated, previous to the task creation.
    """

    @classmethod
    def from_task_cls(cls, task_class, classdict):
        """
        Creating the task definition for a task defined as a class or orchestration decorated task
        This is getting called from the creation of the class itself (by the meta-class)
        """

        # collecting the definitions of the inherited task classes (if any).
        base_task_definitions = get_base_task_definitions(task_class)

        return TaskDefinition(
            classdict=classdict,
            base_task_definitions=base_task_definitions,
            task_passport=TaskPassport.from_task_cls(task_class),
            defaults=classdict.get("defaults", None),
            task_decorator=task_class.task_decorator,
            source_code=TaskSourceCode.from_task_class(task_class),
        )

    @classmethod
    def from_task_decorator(cls, task_decorator):
        # type: (TaskDecorator) -> TaskDefinition
        """
        Creating the task definition from a decorated function.
        This requires the information collected on the function signature
        """
        return TaskDefinition(
            task_passport=task_decorator.task_passport,
            defaults=task_decorator.task_defaults,
            task_decorator=task_decorator,
            source_code=TaskSourceCode.from_callable(
                task_decorator.original_class_or_func
            ),
        )

    def __init__(
        self,
        task_passport,  # type: TaskPassport
        classdict=None,  # type: Optional[Dict[str, Any]]
        base_task_definitions=None,  # type: Optional[List[TaskDefinition]]
        defaults=None,  # type: Optional[Dict[ParameterDefinition, Any]]
        task_decorator=None,  # type: Optional[TaskDecorator]
        source_code=None,  # type: Optional[TaskSourceCode]
        external_parameters=None,  # type: Optional[Parameters]
        task_definition_uid=None,  # type: Optional[UUID]
    ):
        super(TaskDefinition, self).__init__()

        self.hidden = False

        self.task_passport = task_passport
        self.source_code = source_code
        self.task_decorator = task_decorator
        self.base_task_definitions = (
            base_task_definitions or []
        )  # type: List[ TaskDefinition]

        # TODO: maybe use properties or other way to delegate those...
        self.full_task_family = self.task_passport.full_task_family
        self.full_task_family_short = self.task_passport.full_task_family_short
        self.task_family = self.task_passport.task_family
        self.task_config_section = self.task_passport.task_config_section

        # all the attributes that points to_Parameter
        self.task_param_defs = dict()  # type: Dict[str, ParameterDefinition]

        # the defaults attribute
        self.defaults = dict()  # type: Dict[ParameterDefinition, Any]

        self.task_param_defs = self._calculate_task_class_values(
            classdict, external_parameters
        )
        # if we have output params in function arguments, like   f(some_p=parameter.output)
        # the new function can not return the result of return
        self.single_result_output = self._is_result_single_output(self.task_param_defs)

        self.param_defaults = {
            p.name: p.default
            for p in self.task_param_defs.values()
            if is_defined(p.default)
        }

        # TODO: consider joining with task_config
        # TODO: calculate defaults value as _ConfigStore and merge using standard mechanism
        self.defaults = self._calculate_task_defaults(defaults)
        self.task_defaults_config_store = parse_and_build_config_store(
            source=self.task_passport.format_source_name("task.defaults"),
            config_values=self.defaults,
            priority=ConfigValuePriority.FALLBACK,
        )

        self.task_signature_extra = {}
        if config.getboolean("task_build", "sign_with_full_qualified_name"):
            self.task_signature_extra["full_task_family"] = self.full_task_family
        if config.getboolean("task_build", "sign_with_task_code"):
            self.task_signature_extra["task_code_hash"] = user_friendly_signature(
                self.source_code.task_source_code
            )

        if task_definition_uid:
            self.task_definition_uid = task_definition_uid
        else:
            self.task_definition_uid = get_uuid()

    def _calculate_task_class_values(self, classdict, external_parameters):
        # type: (Optional[Dict],  Optional[Parameters]) -> Dict[str, ParameterDefinition]
        # reflect inherited attributes
        params = dict()
        # params will contain definition of param, even it's was overrided by the parent task
        for base_schema in self.base_task_definitions:
            params = combine_mappings(params, base_schema.task_param_defs)

        # let update params with new class attributes
        self._update_params_from_attributes(classdict, params)

        # this is the place we add parameters from function definition
        if self.task_decorator is not None:
            func_params_builder = DecoratedCallableParamBuilder(
                base_params=params, task_decorator=self.task_decorator
            )

            func_params_builder.build_func_params()
            params_dict = dict(func_params_builder.decorator_kwargs_params)
            params_dict.update(func_params_builder.func_spec_params)
            params_dict.update(func_params_builder.result_params)

            self._update_params_from_attributes(params_dict, params)

        if external_parameters:
            params.update(
                {param.name: param for param in external_parameters.get_params()}
            )

        updated_params = {}
        for name, param in six.iteritems(params):
            # add parameters config
            param_with_owner = param.evolve_with_owner(task_definition=self, name=name)

            # updated the owner in the external parameters
            param_value = external_parameters and external_parameters.get_param_value(
                name
            )
            if param_value:
                param_value.parameter = param_with_owner

            updated_params[name] = param_with_owner

        params = _ordered_params(updated_params)
        return params

    def _calculate_task_defaults(self, defaults):
        # type: (...)->  Dict[str, Any]
        base_defaults = dict()
        for base_schema in self.base_task_definitions:
            base_defaults = combine_mappings(base_defaults, base_schema.defaults)

        return combine_mappings(base_defaults, defaults)

    def _update_params_from_attributes(self, classdict, params):
        class_values = dict()
        if not classdict:
            return

        for a_name, a_obj in iteritems(classdict):
            context = "%s.%s" % (self.task_family, a_name)
            try:
                if isinstance(a_obj, ParameterFactory):
                    params[a_name] = a_obj.build_parameter(context)
                elif isinstance(a_obj, ParameterDefinition):
                    params[a_name] = a_obj
                else:
                    class_values[a_name] = a_obj
            except Exception:
                logger.warning("Failed to process %s" % context)
                raise

        # now, if we have overloads in code ( calculated in task_definition):
        # class T(BaseT):
        #     some_base_t_property = new_value
        for p_name, p_val in iteritems(class_values):
            if p_name not in params:
                continue
            params[p_name] = params[p_name].modify(default=p_val)

    def _is_result_single_output(self, params):
        """
        check that task has only one output and it's output is result
         (there can be sub outputs that are part of result)
        """
        result = params.get(RESULT_PARAM)
        if not result:
            return False
        names = result.names if isinstance(result, FuncResultParameter) else []
        for param_def in self.task_param_defs.values():
            if param_def.system or param_def.kind != _ParameterKind.task_output:
                continue
            if param_def.name in [RESULT_PARAM, "task_band"]:
                continue
            if param_def.name in names:
                continue
            return False
        return True

    def __str__(self):
        return "TaskDefinition(%s)" % self.full_task_family

    def run_name(self):
        if self.task_decorator:
            return "%s()" % self.full_task_family
        return "%s.run()" % self.full_task_family


def get_base_task_definitions(task_class):
    """Accessing the task_definitions of the parents classes of the given `task_class`"""
    task_definitions = []

    for cls in reversed(task_class.__bases__):  # type: TaskMetaclass
        if not hasattr(cls, "task_definition"):
            logger.debug(
                "you should inherit from Task objects only: %s -> %s ", task_class, cls
            )
            continue
        task_definitions.append(cls.task_definition)
    return task_definitions



import json
import logging
import typing

from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd

from pandas.core.dtypes.common import is_bool_dtype, is_numeric_dtype, is_string_dtype

from dbnd._core.tracking.schemas.column_stats import (
    ColumnStatsArgs,
    get_column_stats_by_col_name,
)


if typing.TYPE_CHECKING:
    from dbnd._core.tracking.log_data_request import LogDataRequest
    from targets.value_meta import ValueMetaConf

logger = logging.getLogger(__name__)


class PandasHistograms(object):
    """
    calculates histograms and stats on pandas dataframe.
    """

    def __init__(self, df, meta_conf):
        # type: (pd.DataFrame, ValueMetaConf) -> None
        self.df = df
        self.meta_conf = meta_conf

    def get_histograms_and_stats(
        self,
    ) -> Tuple[List[ColumnStatsArgs], Dict[str, List[List]]]:
        columns_stats, histograms = [], {}

        if self.meta_conf.log_stats:
            columns_stats = self._calculate_stats(self.df)

        if self.meta_conf.log_histograms:
            hist_column_names = self._get_column_names_from_request(
                self.df, self.meta_conf.log_histograms
            )
            df_histograms = self.df.filter(hist_column_names)
            if pd.__version__ > "1":
                # return_type="expand" and orient="list" are used to stabilize produced
                # histograms data structure across pandas v0 & v1
                histograms = df_histograms.apply(
                    self._calculate_histograms,
                    result_type="expand",
                    args=(columns_stats,),
                ).to_dict(orient="list")
            else:
                histograms = df_histograms.apply(
                    self._calculate_histograms, args=(columns_stats,)
                ).to_dict()
                histograms = {k: list(v) for k, v in histograms.items()}

        return columns_stats, histograms

    def _get_column_names_from_request(self, df, data_request):
        # type: (pd.DataFrame, LogDataRequest) -> List[str]
        column_names = list(data_request.include_columns)
        for column_name, column_type in df.dtypes.items():
            if data_request.include_all_string and is_string_dtype(column_type):
                column_names.append(column_name)
            elif data_request.include_all_boolean and is_bool_dtype(column_type):
                column_names.append(column_name)
            elif data_request.include_all_numeric and is_numeric_dtype(column_type):
                column_names.append(column_name)

        column_names = [
            column
            for column in column_names
            if column not in data_request.exclude_columns
        ]
        return column_names

    def _calculate_stats(self, df):
        # type: (pd.DataFrame) -> Dict[str, Dict]
        stats_column_names = self._get_column_names_from_request(
            df, self.meta_conf.log_stats
        )
        df = df.filter(stats_column_names)
        try:
            stats = df.describe(include="all").to_json()
        except Exception as e:
            logger.warning("Failed to describe df: %s", e)
            stats = df.explode().describe(include="all").to_json()
        stats = json.loads(stats)
        stats = self._remove_none_values(stats)
        columns_stats: List[ColumnStatsArgs] = []
        for column_name in stats.keys():
            # Calculate ColumnStatsArgs metrics
            records_count = df[column_name].size
            null_count = np.count_nonzero(pd.isnull(df[column_name]))
            try:
                distinct_count = len(df[column_name].unique())
            except Exception:
                logger.warning("Failed to determine column type for: %s.", column_name)
                try:
                    distinct_count = len(df[column_name].astype("str").unique())
                except Exception:
                    # Support pandas >= v1.0
                    distinct_count = len(df[column_name].astype("string").unique())

            column_type = self._get_column_type(df[column_name])
            column_stats = ColumnStatsArgs(
                column_name=column_name,
                column_type=column_type,
                records_count=records_count,
                null_count=null_count,
                distinct_count=distinct_count,
                most_freq_value=stats[column_name].get("top"),
                most_freq_value_count=stats[column_name].get("freq"),
                unique_count=stats[column_name].get("unique"),
                mean_value=stats[column_name].get("mean"),
                min_value=stats[column_name].get("min"),
                max_value=stats[column_name].get("max"),
                std_value=stats[column_name].get("std"),
                quartile_1=stats[column_name].get("25%"),
                quartile_2=stats[column_name].get("50%"),
                quartile_3=stats[column_name].get("75%"),
            )
            columns_stats.append(column_stats)

        return columns_stats

    def _remove_none_values(self, input_dict):
        """remove none values from dict recursively"""
        for key, value in list(input_dict.items()):
            if value is None:
                input_dict.pop(key)
            elif isinstance(value, dict):
                self._remove_none_values(value)
        return input_dict

    def _get_column_type(self, column):
        # type: (pd.Series) -> str
        first_index = column.first_valid_index()
        if first_index is None:
            return column.dtype.name
        first_value = column.at[first_index]
        return type(first_value).__name__

    def _calculate_histograms(self, df_column, columns_stats):
        # type: (pd.Series, List[ColumnStatsArgs]) -> Optional[Tuple[List, List]]
        try:
            if len(df_column) == 0:
                return

            column_type = self._get_column_type(df_column)
            if "1" < pd.__version__ < "1.2":
                # handle  'Float32', 'Float64', 'Int8', 'Int16', 'Int32', 'Int64', 'UInt8', 'UInt16', 'UInt32', 'UInt64'
                column_type = column_type.lower()
            if is_string_dtype(column_type) or is_bool_dtype(column_type):
                counts = df_column.value_counts()  # type: pd.Series
                column_stats = get_column_stats_by_col_name(
                    columns_stats, df_column.name
                )
                if column_stats and column_stats.null_count:
                    null_column = pd.Series([column_stats.null_count], index=[None])
                    counts = counts.append(null_column)
                    counts = counts.sort_values(ascending=False)
                if len(counts) > 50:
                    counts, tail = counts[:49], counts[49:]
                    tail_sum = pd.Series([tail.sum()], index=["_others"])
                    counts = counts.append(tail_sum)
                values = counts.index
            elif is_numeric_dtype(column_type):
                df_column = df_column.dropna()
                counts, values = np.histogram(
                    df_column, bins=20
                )  # Tuple[np.array, np.array]
            else:
                return

            return counts.tolist(), values.tolist()
        except Exception:
            logger.exception(
                "log_histogram: Something went wrong for column '%s'", df_column.name
            )



"""Module dbnd.tasks.basics contains some pipeline examples to start with."""
from dbnd.tasks.basics.sanity import dbnd_sanity_check
from dbnd.tasks.basics.shell import bash_cmd, bash_script
from dbnd.tasks.basics.simplest import SimplestPipeline, SimplestTask


__all__ = [
    "dbnd_sanity_check",
    "bash_cmd",
    "bash_script",
    "SimplestPipeline",
    "SimplestTask",
]



from dbnd._core.task.config import Config
from dbnd._core.task.data_source_task import DataSourceTask, data_combine, data_source
from dbnd._core.task.pipeline_task import PipelineTask
from dbnd._core.task.python_task import PythonTask
from dbnd._core.task.task import Task
from dbnd.tasks import basics
from dbnd.tasks.doctor.check import dbnd_doctor


__all__ = [
    "Task",
    "Config",
    "data_combine",
    "data_source",
    "DataSourceTask",
    "PipelineTask",
    "PythonTask",
    "dbnd_doctor",
    "basics",
]



import logging
import typing

from collections.abc import Callable
from datetime import timedelta
from itertools import chain
from math import floor, log10

import six

from more_itertools import collapse, partition

from dbnd._core.configuration.environ_config import in_airflow_tracking_mode
from dbnd._core.constants import SystemTaskName, TaskRunState
from dbnd._core.current import is_verbose
from dbnd._core.tracking.backends import TrackingStore
from dbnd._core.tracking.schemas.metrics import Metric
from dbnd._core.utils.basics.text_banner import safe_tabulate
from dbnd._core.utils.timezone import utcnow
from dbnd._vendor.ascii_graph import Pyasciigraph


try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

if typing.TYPE_CHECKING:
    from typing import Any, Dict, Iterable, List, Tuple

    from dbnd._core.task_run.task_run import TaskRun
    from targets.value_meta import ValueMeta

logger = logging.getLogger(__name__)


def is_describe_stat(item):
    name, _ = item
    return name in ["min_value", "quartile_1", "quartile_2", "quartile_3", "max_value"]


def two_columns_table(stats, split_by):
    # type: (Dict, Callable[Tuple[Any,Any], bool]) -> str
    """
    Split the items by a filter function "split_by" and merge
    Example:
        two_columns_table({1: "a", 2: "b", 3: "c", 4: "d"}, is_odd)
        [ 2 , b , 1 , a ]
        [ 3 , d , 3 , c ]
    """
    splatted = partition(split_by, stats.items())
    flatten_merged_rows = map(collapse, zip_longest(*splatted))
    return safe_tabulate(flatten_merged_rows, headers=())


class ConsoleStore(TrackingStore):
    def __init__(self, *args, **kwargs):
        super(ConsoleStore, self).__init__(*args, **kwargs)
        self.max_log_value_len = 50
        self.verbose = is_verbose()
        self.ascii_graph = Pyasciigraph()
        self._is_in_airflow_tracking_mode = in_airflow_tracking_mode()

    def init_run(self, run):
        if run.is_orchestration:
            logger.info(
                run.describe.run_banner(
                    "Running Databand v%s" % run.context.task_run_env.databand_version,
                    color="cyan",
                    show_run_info=True,
                )
            )

            if run.context.name == "interactive":
                from dbnd._core.tools import ipython

                ipython.show_run_url(run.run_url)

    def set_task_reused(self, task_run):
        task = task_run.task
        logger.info(
            task.ctrl.visualiser.banner(
                "Task %s has been completed already!" % task.task_id,
                "magenta",
                task_run=task_run,
            )
        )

    def set_task_run_state(self, task_run, state, error=None, timestamp=None):
        super(ConsoleStore, self).set_task_run_state(task_run=task_run, state=state)

        task = task_run.task

        if self._is_in_airflow_tracking_mode:
            if state == TaskRunState.RUNNING:
                logger.info(
                    "Tracking %s task at %s", task.task_name, task_run.task_tracker_url
                )
            return

        # optimize, don't print success banner for fast running tasks
        start_time = task_run.start_time or utcnow()
        quick_task = task_run.finished_time and (
            task_run.finished_time - start_time
        ) < timedelta(seconds=5)

        show_simple_log = not self.verbose and (
            task_run.task.task_is_system or quick_task
        )

        level = logging.INFO
        color = "cyan"
        task_friendly_id = task_run.task_af_id
        task_id_str = "%s(%s)" % (task.task_id, task_friendly_id)
        if state in [TaskRunState.RUNNING, TaskRunState.QUEUED]:
            task_msg = "Running task %s" % task_id_str

        elif state == TaskRunState.SUCCESS:
            task_msg = "Task %s has been completed!" % task_id_str
            color = "green"

        elif state == TaskRunState.FAILED:
            task_msg = "Task %s has failed!" % task_id_str
            color = "red"
            level = logging.WARNING
            if task_run.task.task_name in SystemTaskName.driver_and_submitter:
                show_simple_log = True

        elif state == TaskRunState.CANCELLED:
            task_msg = "Task %s has been canceled!" % task_id_str
            color = "red"
            level = logging.WARNING
            show_simple_log = True

        else:
            task_msg = "Task %s moved to %s state" % (task_friendly_id, state)

        if show_simple_log:
            logger.log(level, task_msg)
            return

        try:
            logger.log(
                level, task.ctrl.visualiser.banner(task_msg, color, task_run=task_run)
            )
        except Exception as ex:
            logger.log(level, "%s , failed to create banner: %s" % (task_msg, ex))

    def add_task_runs(self, run, task_runs):
        pass

    def save_external_links(self, task_run, external_links_dict):
        task = task_run.task
        attempt_uid = task_run.task_run_attempt_uid

        logger.info(
            "%s %s has registered URLs to external resources: %s",
            task,
            attempt_uid,
            ", ".join("%s: %s" % (k, v) for k, v in six.iteritems(external_links_dict)),
        )

    def is_ready(self):
        return True

    def log_histograms(self, task_run, key, value_meta, timestamp):
        # type: (TaskRun, str, ValueMeta, datetime) -> None

        for graph in self.get_console_histograms(key, value_meta):
            for line in graph:
                # we add one line at a time using the logger to keep the output aligned together
                logger.info(line)

        if not self.verbose:
            return
        if value_meta.histogram_system_metrics:
            for key, value in value_meta.histogram_system_metrics.items():
                if isinstance(value, float):
                    value = round(value, 3)
                logger.info("%s: %s", key, value)

    def get_console_histograms(self, key, value_meta):
        # type: (str, ValueMeta) -> Iterable[Iterable[str]]
        """
        Iterate over histograms and for each its build a all the lines to print to the console
        """
        for column_name, hist_values in value_meta.histograms.items():
            console_graph = self.ascii_graph.graph(
                label=f"Histogram logged: {key}.{column_name}",
                data=list(zip(*reversed(hist_values))),
            )

            # not will exists, but when it does - we need to add the stats as a table to the console
            if value_meta.columns_stats:
                column_stats = value_meta.get_column_stats_by_col_name(column_name)
                if column_stats:
                    stats_table = two_columns_table(
                        column_stats.as_dict(), is_describe_stat
                    )
                    # channing the printable graph and status table
                    console_graph = chain(console_graph, stats_table.splitlines())

            # separation line between one histogram and the other
            console_graph = chain(console_graph, [""])
            yield console_graph

    @staticmethod
    def _format_value(value):
        """
        Convert value to string.
        Additionally forces float values to NOT use scientific notation.
        """
        if isinstance(value, float):
            value_width = 1 if value == 0.0 else abs(floor(log10(abs(value))))
            # default python precision threshold to switch to scientific representation of floats
            if value > 1 or value_width < 4:
                return str(value)
            return "{:{width}.{prec}f}".format(
                value, width=value_width, prec=value_width + 4
            )
        return str(value)

    def log_metrics(self, task_run, metrics):
        # type: (TaskRun, List[Metric]) -> None
        if len(metrics) == 1:
            m = metrics[0]
            logger.info(
                "Task {name} metric: {key}={value}".format(
                    name=task_run.task_af_id,
                    key=m.key,
                    value=self._format_value(m.value),
                )
            )
        else:
            metrics_str = ", ".join(
                ["{}={}".format(m.key, self._format_value(m.value)) for m in metrics]
            )
            logger.info("Task %s metrics: %s", task_run.task_af_id, metrics_str)
